unit UnicodeUnit;

interface

//        UTF-8    UTF-32   -1,    
function GetCharUTF8(var P: PByte; var L: integer): integer;

//    UTF-32    (    !)
function LowerUTF32(ACode: integer): integer;

//   UTF-8  (case-insensitive    )
function CompareUTF8IgnoreCase(L1: integer; P1: PByte; L2: integer; P2: PByte): integer;


implementation

uses SysUtils;

//    UTF-8
type
  TParseItem = record
    Mask: integer;
    Count: integer;
  end;


var
//    UTF-8
  ParseTable: array[0..255] of TParseItem;

//       
  LowerTable: array[0..65535] of integer;

function CompareUTF8IgnoreCase(L1: integer; P1: PByte; L2: integer; P2: PByte): integer;
var
  C1, C2: integer;
begin

  repeat

    if (L1 = 0) and (L2 = 0) then begin
      result := 0;
      exit;
    end;

    //    
    C1 := GetCharUTF8(P1, L1); if C1 < 0 then begin result := -1; exit; end;
    C2 := GetCharUTF8(P2, L2); if C2 < 0 then begin result := +1; exit; end;

    //        
    if C1 < 65536 then C1 := LowerTable[C1];
    if C2 < 65536 then C2 := LowerTable[C2];


    if C1 < C2 then begin
      result := -1;
      exit;
    end else if C1 > C2 then begin
      result := +1;
      exit;
    end;

  until false;

end;


function LowerUTF32(ACode: integer): integer;
begin

  case ACode of

    1105, 1025: // , 
      result := 1077; // 

    1040..1071:
      result := ACode + 32; // 

    Ord('A')..Ord('Z'):
      result := ACode + 32; // 

  else
    result := ACode;

  end;
end;


function GetCharUTF8(var P: PByte; var L: integer): integer;
var
  B: byte;
  I: integer;

begin

  if L > 0 then begin

    B := P^; Inc(P); Dec(L);

    with ParseTable[B] do
      if Count > 0 then begin

        //     
        result := B and Mask;

        //     
        for I := 1 to Count - 1 do

          if L > 0 then begin
            result := (result shl 6) or (P^ and $3F);
            Inc(P); Dec(L);
          end else begin
            result := -1;
            exit;
          end;

        exit;
      end;

  end;

  result := -1;
  
end;

var
  I: integer;

initialization

//    UTF-8
  for I := 0 to 255 do
    with ParseTable[I] do
      if I <= 127 then begin
        Mask := $7F;
        Count := 1;
      end else if I >= 248 then begin
        Mask := 0;
        Count := 0;
      end else if I >= 240  then begin
        Mask := 7;
        Count := 4;
      end else if I >= 224  then begin
        Mask := 15;
        Count := 3;
      end else if I >= 192  then begin
        Mask := $1F;
        Count := 2;
      end else begin
        Mask := 0;
        Count := 0;
      end;

  //  Lower
  for I := 0 to 65535 do
    LowerTable[I] := LowerUTF32(I);


end.